#!/usr/bin/python


# The MIT License (MIT)
#
# Copyright (c) 2014 Dave Parsons
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys

# Setup imports depending on whether IronPython or CPython
if sys.platform == 'cli':
    from Microsoft.Win32 import Registry
if sys.platform == 'win32':
    import os
    from _winreg import *
else:
    import os


def allindices(string, sub, listindex, offset=0):
    i = string.find(sub, offset)
    while i >= 0:
        listindex.append(i)
        i = string.find(sub, i + 1)
    return listindex


def rot13(s):
    chars = 'AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz'
    trans = chars[26:] + chars[:26]
    rot_char = lambda c: trans[chars.find(c)] if chars.find(c) > -1 else c
    return ''.join(rot_char(c) for c in s)


def patchsmc(f, lks, osk0, osk1):
    # Key Data Structure
    # 0x00/00 0x04/04 int  Key name (byte reversed e.g. #KEY is #YEK)
    # 0x04/04 0x01/01 byte Length of returned data
    # 0x05/05 0x04/04 int  Data type of returned data (byte reversed e.g. ui32 is 23iu)
    # 0x09/09 0x01/01 byte Flag R/W
    # 0x0a/10 0x06/06 byte Padding
    # 0x10/16 0x08/08 ptr  Internal VMware routine
    # 0x18/24 0x30/48 byte Data

    # Get memcpy routine address from +LKS key
    f.seek(lks + 16)
    lks_cpy = f.read(8)
    print '+LKS Memcpy: 0x' + lks_cpy.encode('hex')

    # Copy to OSK0 key
    f.seek(osk0 + 16)
    osk0_cpy = f.read(8)
    print 'OSK0 Memcpy Original: 0x' + osk0_cpy.encode('hex')
    f.seek(osk0 + 16)
    f.write(lks_cpy)
    f.seek(osk0 + 16)
    osk0_cpy = f.read(8)
    print 'OSK0 Memcpy Patched:  0x' + osk0_cpy.encode('hex')

    # Copy to OSK1 key
    f.seek(osk1 + 16)
    osk1_cpy = f.read(8)
    print 'OSK1 Memcpy Original: 0x' + osk1_cpy.encode('hex')
    f.seek(osk1 + 16)
    f.write(lks_cpy)
    f.seek(osk1 + 16)
    osk1_cpy = f.read(8)
    print 'OSK1 Memcpy Patched:  0x' + osk1_cpy.encode('hex')

    # Copy Haiku to data section
    osk0_text = rot13('bheuneqjbexolgurfrjbeqfthneqrqcy')
    osk1_text = rot13('rnfrqbagfgrny(p)NccyrPbzchgreVap')
    f.seek(osk0 + 24)
    f.write(osk0_text)
    f.seek(osk1 + 24)
    f.write(osk1_text)
    print 'Haiku written to OSK0/1 Data'


def patchvmx(name):

    # Patch file
    f = open(name, 'r+b')
    print 'Patching: ' + name

    # Read file into string variable
    vmx = f.read()

    # Setup hex strings
    lks_key = '\x53\x4B\x4C\x2B\x01\x67\x61\x6C\x66'
    osk0_key = '\x30\x4B\x53\x4F\x20\x2A\x38\x68\x63'
    osk1_key = '\x31\x4B\x53\x4F\x20\x2A\x38\x68\x63'

    # Find +LKS flag
    lks_0 = vmx.find(lks_key)
    lks_1 = vmx.rfind(lks_key)

    # Find OSK0 ch8*
    osk0_0 = vmx.find(osk0_key)
    osk0_1 = vmx.rfind(osk0_key)

    # Find OSK1 ch8*
    osk1_0 = vmx.find(osk1_key)
    osk1_1 = vmx.rfind(osk1_key)

    # Print details & patch file
    print '+LKS 0: ' + hex(lks_0)
    print 'OSK0 0: ' + hex(osk0_0)
    print 'OSK1 0: ' + hex(osk1_0)
    patchsmc(f, lks_0, osk0_0, osk1_0)

    print '+LKS 1: ' + hex(lks_1)
    print 'OSK0 1: ' + hex(osk0_1)
    print 'OSK1 1: ' + hex(osk1_1)
    patchsmc(f, lks_1, osk0_1, osk1_1)

    # Tidy up
    f.flush()
    f.close()
    print 'Patched: ' + name


def patchbase(name):

    # Patch file
    print 'Patching: ' + name
    f = open(name, 'r+b')

    # Entry to search for in GOS table
    darwin = (
        '\x10\x00\x00\x00\x10\x00\x00\x00'
        '\x02\x00\x00\x00\x00\x00\x00\x00'
        '\x00\x00\x00\x00\x00\x00\x00\x00'
        '\x00\x00\x00\x00\x00\x00\x00\x00'
        '\xBE'
    )

    # Read file into string variable
    base = f.read()

    # Find all occurences of darwin table entry
    offsets = []
    allindices(base, darwin, offsets, 0)
    print offsets

    # Loop thorugh each entry and set top bit
    # 0xBE --> 0xBF
    for offset in offsets:
        f.seek(offset + 32)
        flag = f.read(1)
        if flag.encode('hex') == 'be':
            f.seek(offset + 32)
            f.write('\xBF')
            print 'Patched flag: 0x' + str(offset)
        else:
            print 'Unknown flag :' + flag.encode('hex')

    # Tidy up
    f.flush()
    f.close()
    print 'Patched: ' + name


def main():
    # Work around absent Platform module on VMkernel
    if sys.platform == 'cli':
        osname = 'windows-cli'
    else:
        # import os
        if os.name == 'nt':
            osname = 'windows'
        else:
            osname = os.uname()[0].lower()

    # Setup default paths
    if osname == 'darwin':
        vmx = 'vmware-vmx'
        vmx_debug = 'vmware-vmx-debug'
        vmx_stats = 'vmware-vmx-stats'
        vmwarebase = ''
        vmx_path = '/Applications/VMware Fusion.app/Contents/Library/'
        vmwarebase_path = ''

    elif osname == 'linux':
        vmx = 'vmware-vmx'
        vmx_debug = 'vmware-vmx-debug'
        vmx_stats = 'vmware-vmx-stats'
        vmwarebase = 'libvmwarebase.so.0'
        vmx_path = '/usr/lib/vmware/bin/'
        vmwarebase_path = '/usr/lib/vmware/lib/libvmwarebase.so.0/'

    # elif osname == 'vmkernel':
    #     vmx = 'vmx'
    #     vmx_debug = 'vmx-debug'
    #     vmx_stats = 'vmx-stats'
    #     vmwarebase = ''
    #     vmx_path = '/unlocker/'
    #     vmwarebase_path = ''

    elif osname == 'windows':
        vmx = 'vmware-vmx.exe'
        vmx_debug = 'vmware-vmx-debug.exe'
        vmx_stats = 'vmware-vmx-stats.exe'
        vmwarebase = 'vmwarebase.dll'
        reg = ConnectRegistry(None, HKEY_LOCAL_MACHINE)
        key = OpenKey(reg, r'SOFTWARE\Wow6432Node\VMware, Inc.\VMware Workstation')
        vmwarebase_path = QueryValueEx(key, 'InstallPath')[0]
        vmx_path = QueryValueEx(key, 'InstallPath64')[0]

    elif osname == 'windows-cli':
        vmx = 'vmware-vmx.exe'
        vmx_debug = 'vmware-vmx-debug.exe'
        vmx_stats = 'vmware-vmx-stats.exe'
        vmwarebase = 'vmwarebase.dll'
        reg = Registry.LocalMachine.OpenSubKey(r'SOFTWARE\Wow6432Node\VMware, Inc.\VMware Workstation')
        vmwarebase_path = reg.GetValue('InstallPath')
        vmx_path = reg.GetValue('InstallPath64')

    else:
        print('Unknown Operating System: ' + osname)
        return

    # VMX path from command line
    if len(sys.argv) >= 2:
        vmx_path = sys.argv[1]

    # VMWAREBASE path fron command line
    if len(sys.argv) >= 3:
        vmwarebase_path = sys.argv[2]

    #  Patch the vnx executiables skipping stats version for Player
    patchvmx(vmx_path + vmx)
    patchvmx(vmx_path + vmx_debug)
    try:
        patchvmx(vmx_path + vmx_stats)
    except IOError:
        pass

    # Patch vmwarebase DLL/SO for Workstation and Player
    # Not required on Fusionor ESXi as table already has correct flags
    if vmwarebase_path != '':
        patchbase(vmwarebase_path + vmwarebase)
    else:
        print 'Patching vmwarebase is not required on this system'


if __name__ == '__main__':
    main()
